Unit dualQry; { Small dataset derived class }

{
  This component was derived from the 16 bit TUQBE class developed by David Berneda, which
  has always been free. Grant Carpenter made an attempt to update it to 32 bits but some of
  its functionality was lost.  I modified the original 16 bit control to regain functionality
  but since I find that I use a mix of QBE and SQL queries, I thought it would be nice to
  have a component which would handle either one.  The BDE only requires that you tell it
  what query language it should expect.  It will cheerfully execute a query in either format.
  It has been substantially modified by me and thus carries the Morphic Molecules logo.
  However it is free for any use, commercial or otherwise.
  It will be supported to a limited extent by email (everett@txdirect.net)
  DISCLAIMER : I take no responsibility for any damage, loss of goods, services,etc.
  which might be caused by the use of this control by you or anyone using software
  in which you have included this control.

  MMEverett 
  everett@txdirect.net

}



{ Modified original notes. }
(*
  Answer table name can be specified with or without alias.
  Eg: :dbdemos:void.db
      void.db -- you must have set up drive and types correctly.
      c:\mydb\void.db

  Assumed answer table Driver: DBASE
  THE OLD COMPONENT USED PARADOX as a DEFAULT.
  The new property: AnswerType allows to PARADOX,DBASE and ASCII answer table types

  If the answer table exists, an atempt to delete it is made before copying
  the result cursor.
  The physical answer table must be opened in exclusive mode and all related
  family files are erased together with the table. That is, if something else has
  this table open, the query will error out.
  Unfortunately, if the query does not return a table; e.g. in update, insert, or delete
  queries, you will get an error message saying that I could not set a cursor, although the
  query has executed just fine.  You can trap for this with an on EDatabaseError in a 
  try..except clause, although you will still get the error when running under Delphi.

  Other functions in this unit:

**  Function GetAliasPath(Const Alias:String):String;
      Returns the path for the "alias" or a empty string if not found.
      Eg: ('dbdemos') returns 'c:\delphi\demos\data'

**  Function GetDBTablePath(Const TableName:String):String;
      Returns the TableName with path instead of alias (if it has an alias).
      Eg: (':dbdemos:customer.db') returns 'c:\delphi\demos\data\customer.db'

*)

interface

Uses Classes,DBTables,DB,SysUtils,DBConsts,LibConst,DBIprocs,DBIErrs,DBITypes,DsgnIntf,Dialogs,Forms;

Const MaxParam   = 10;  { max number of query parameters }
      MaxParamLen=255; { max length of a substituted param }

Type
  tQueryType = (qtQBE,qtSQL);
Type
     TDualQuery=Class(TDBDataSet)
     private
       FFileName : TFileName;
       FAnswerTable:String;
       FAnswerType:TTableType;
       FQueryType: tQueryType;
       FBlankasZero,
       FAuxTables,
       FRequestLive:Boolean;
       eQueryLang : dbiQryLang;
     protected
       function CreateHandle: HDBICur; override;
     public
       FQuery: TStrings;
       FParams : TStrings;
       NumParam:Integer;
       Param,Subst:Array[0..MaxParam] of String[MaxParamLen];
       procedure SetQuery(QLines: TStrings);
       Constructor Create(AOwner:TComponent); override;
       destructor Destroy; override;
       Procedure AddParam(Const tmpParam,tmpSubst:String; doReplace: boolean);
       Procedure DeleteParam(Const tmpParam:String; warn: boolean);
       Procedure ClearParams;
       procedure SetQueryLang( which : tQueryType);
       Function  GetAliasPath(Const Alias:String):String;
       Function  GetDBTablePath(Const TableName:String):String;
       Function  ReplaceString(s:String):String;
       Function  LoadFromFile( fname : string):boolean;
       Function  SaveToFile( fname : string):boolean;
     published
       property QLines: TStrings read FQuery write SetQuery;
       property QParams: TStrings read FParams write FParams;
       property AnswerTable: String read FAnswerTable write FAnswerTable;
       property RequestLive: Boolean read FRequestLive write FRequestLive;
       property BlankasZero: Boolean read FBlankasZero write FBlankAsZero;
       property AuxTables: Boolean read FAuxTables write FAuxTables;
       property AnswerType:TTableType read FAnswerType write FAnswerType;
       property QueryLanguage: tQueryType read fQueryType write setQueryLang;
       property Filename : TFileName read fFileName write fFileName;
     End;

Procedure Register;

implementation


Constructor TDualQuery.Create(AOwner:TComponent);
Begin
  inherited Create(AOwner);
  FQuery := TStringList.Create;
  FParams := TStringList.create;

  NumParam:=0;
  FAnswerType:=ttDBase; { by default, DBase answer tables }
  QueryLanguage := qtQBE;
end;

destructor TDualQuery.Destroy;
Begin
  FQuery.Free;
  FParams.free;
  inherited Destroy;
End;

Function  TDualQuery.LoadFromFile( fname : string): boolean;
begin
  if not fileExists(fname) then begin
    result := false;
    exit;
  end;
  FQuery.loadFromFile(fname);
  filename := '';
  result := true;
end;

Function  TDualQuery.SaveToFile( fname : string): boolean;
begin
  FQuery.saveToFile(fname);
  result := true;
end;


Procedure TDualQuery.setQueryLang( which : TQueryType);
begin
  if which = qtQBE then eQueryLang := qrylangQBE
  else eQueryLang := qrylangSQL;
end;



Procedure TDualQuery.ClearParams;
Begin
  QParams.clear;
  fParams.clear;
End;

Procedure TDualQuery.DeleteParam(Const tmpParam:String; warn: boolean);
var
  qstr,par :string;
  xx,ii,k : integer;
Begin
  k := fParams.count;
  for ii := 0 to k-1 do begin
    qstr := fparams[ii];
    xx := pos(',',qstr) ;
    par := copy(qstr,1,xx-1);
    if par = tmpParam then begin
      if not warn then begin
        fParams.delete(ii);
        exit;
      end { end do replace }
      else begin
        messageDlg('Deleting' + tmpParam ,mtInformation,[mbOK],0);
        exit;
      end; { end don't replace }
    end; { end par=tmpParam}
  end;{ end ii for loop }
end;

Procedure TDualQuery.AddParam(Const tmpParam,tmpSubst:String;doReplace: boolean);
var
  qstr,par :string;
  xx,ii,k : integer;
  alreadythere : boolean;
Begin
  k := fParams.count;
  alreadythere := false;
  for ii := 0 to k-1 do begin
    qstr := fparams[ii];
    xx := pos(',',qstr) ;
    par := copy(qstr,1,xx-1);
    if par = tmpParam then begin
      if doReplace then begin
        fParams[ii] :=  tmpParam + ',' + tmpSubst;
      end { end do replace }
      else begin
        messageDlg('Did not replace parameter ' + tmpParam + ' with ' + tmpSubst + '.',mtInformation,[mbOK],0);
      end; { end don't replace }
      alreadythere := true;
    end; { end par=tmpParam}
  end;{ end ii for loop }
  if not alreadyThere then fparams.add(tmpParam + ',' + tmpSubst);
End;

Function TDualQuery.ReplaceString(s:String):String;
Var
  t,ii,k,xx:Integer;
  qstr,par,subs : string;
  lstr,rstr:string;
Begin
if s = '' then
begin
  result := s;
  exit;
end;
k := fParams.count;
  for t:=0 to k-1 do begin
    qstr := fParams[t];
    if qstr = '' then continue;
    xx := Pos(',',qstr);
    par := copy (qstr,1,xx-1) ;
    subs := copy(qstr,xx+1,length(qstr)-xx);
    Repeat
      ii:=Pos(par,s);
      if ii>0 then
        begin
          lstr := Copy(s,1,ii-1);
          rstr := Copy(s,ii + length(par),length(s)+length(par) + length(subs));
          s := lstr + subs + rstr;
        end;
    Until ii=0;
  end;
  result:=s;
End;

function TDualQuery.CreateHandle: HDBICur;
Var p:HDbiCur;
    Stmt:hDBIStmt;
    {St:Array[0..255] of Char; }
    aBatTblDesc:BATTblDesc;
    r:Longint;
    dbiErr:DBIRESULT;
    NewQuery:TStrings;
    t,d:Integer;
    tmpType:String;
    beforestr,afterstr : string;
Begin
  if fFileName <> '' then begin
    if fileExists(fFileName) then FQuery.loadFromFile(fFileName)
    else begin
      messageDlg('No such file!', mtError, [mbOK],0);
      exit;
    end;
  end;
  NewQuery:=TStringList.Create;
  With FQuery do
  for t:=0 to Count-1 do
  begin
    beforestr := strings[t];
    NewQuery.Add(ReplaceString(Strings[t]));
    afterstr := strings[t];
    d := t;
  end;
  DbiQAlloc( DBHandle, eQueryLang,  Stmt );

  if FRequestLive then Check(dbiSetProp(hDBIObj(Stmt),stmtLIVENESS,longint(wantLive)))
                  Else Check(dbiSetProp(hDBIObj(Stmt),stmtLIVENESS,longint(wantDefault)));
  if FBlankAsZero then Check(dbiSetProp(hDBIObj(Stmt),stmtBLANKS,1));
  if FAuxTables then Check(dbiSetProp(hDBIObj(Stmt),stmtAUXTBLS,1));
  check(dbiQExecDirect(DBHandle,eQueryLang,NewQuery.GetText,@p));
  Check(dbiQFree(Stmt));
  if (FAnswerTable<>'') And Assigned(p) then
  Begin
    Check(DbiSetToBegin(p));
    With aBatTblDesc do
    Begin
      hDB:=DBHandle;
      StrPCopy(szTblName,GetDBTablePath(FAnswerTable));
      Case FAnswerType of
        ttParadox: tmpType:=szParadox;
        ttDbase  : tmpType:=szDbase;
        ttAscii  : tmpType:=szAscii;
      end;
      StrPCopy(szTblType,tmpType);
      szUsername[0]:=#0;
      szPassword[0]:=#0;
    End;
    r:=0;
    dbiErr:=dbiDeleteTable(DBHandle,aBatTblDesc.szTblName,aBatTblDesc.szTblType);
    if dbiErr<>DBIERR_NOSUCHTABLE then Check(dbiErr);
    Check(DbiBatchMove(nil,p,@aBatTblDesc,nil,batchCOPY,0,
                            nil, nil, nil, 0, nil, nil,
                            nil, nil, nil, nil, TRUE, TRUE,
                            r, TRUE));
  End;
  NewQuery.Free;
  result := p;
  
End;

procedure TDualQuery.SetQuery(QLines: TStrings);
begin
  FQuery.Assign(QLines);
end;

Function HasAlias(Const TableName:String):Boolean;
Begin
  Result:=Pos(':',TableName)>0;
End;

Function TDualQuery.GetAliasPath(Const Alias:String):String;
Var AliasList:TStringList;
    i:Longint;
    DBPath:String;
Begin
  Result:='';
  AliasList:=TStringList.Create;
  try
    Session.GetAliasNames(AliasList);
    i:=AliasList.IndexOf(Alias);
    if i<0 then raise EDatabaseError.Create('Alias '+Alias+' doesnt exist')
    else
    Begin
      Session.GetAliasParams(Alias,AliasList);
      DBPath := AliasList.Values['PATH'];
      if DBPath='' then raise EDatabaseError.Create('Alias path from '+Alias+' invalid')
                   else Result:=DBPath;
    end;
  finally
    AliasList.Free;
  end;
End;

Procedure SplitTableName(Const TableName:String; Var Alias,Name:String);
Var p1,p2:Integer;
Begin
  Name:=TableName;
  Alias:='';
  p1:=Pos(':',TableName);
  if p1>0 then
  Begin
    p2:=Pos(':',Copy(TableName,p1+1,255));
    if p2>0 then
    Begin
      Alias:=Copy(TableName,p1+1,p2-1);
      Name:=Copy(TableName,p1+p2+1,255);
    End;
  End;
End;

Function TDualQuery.GetDBTablePath(Const TableName:String):String;
Var Alias,Name:String;
Begin
  if not HasAlias(TableName) then Result:=TableName
  else
  Begin
    SplitTableName(TableName,Alias,Name);
    if Alias<>'' then Result:=GetAliasPath(Alias)+'\'+Name
                 else Result:=TableName;
  End;
End;

Procedure Register;
Begin
  RegisterComponents(LoadStr(srDAccess),[TDualQuery]);
End;

end.
